/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package ws3dproxy.model;

import ws3dproxy.util.Constants;
import java.awt.Color;
import java.awt.Polygon;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;
import ws3dproxy.CommandExecException;
import ws3dproxy.CommandUtility;
import ws3dproxy.WS3DProxy;

/**
 *
 * @author eccastro
 */
public class Creature {

    public Actuator a;
    public MySensors s;
    private WorldMap worldMap;
    private Color color;
    private long mySize;
    private int motorSys; //1- car-like; otherwise, two-wheeled differential steering
    private int hasLeaflet = 0; //false
    private HashMap<Long, Leaflet> myLeaflets = new HashMap<Long, Leaflet>();
    private String name;
    static final Logger logger = Logger.getLogger(Creature.class.getName());
    private String robotIndexID;
    private static Creature instance = null;
    private List<Thing>thingsInVision = new ArrayList<Thing>();

    private Creature(String indexID, String myName, String colorName, double speed, double wheel, double size, double pitch, int motorSys, double fuel, WorldPoint position, int hasLeaflet) {
        robotIndexID = indexID;
        this.color = translateIntoColor(colorName); //used in MindWindow
        this.motorSys = motorSys;
        a = new Actuator();
        s = new MySensors();
        if (speed == 0) {
            speed = Constants.DEFAULT_CREATURE_SPEED;//5 default (old: 1)
        }
        this.mySize = Math.round(size);
        name = myName;
        this.hasLeaflet = hasLeaflet;

//        Constants.GAP = (double) (2 * this.mySize);
//        Constants.SEC = (double) (1.1 * this.mySize); //optimized (not 1.0) due to NPFood size

        this.a.speed = speed;
        this.a.wheel = wheel;
        this.s.pitch = pitch;
        this.s.fuel = fuel;
        this.s.x = position.getX();
        this.s.y = position.getY();

    }

        public static Creature getInstance(String indexID, String myName, String colorName, double speed, double wheel, double size, double pitch, int motorSys, double fuel, WorldPoint position, int hasLeaflet) {
        if (instance == null) {
            instance = new Creature(indexID,  myName, colorName, speed, wheel, size, pitch, motorSys, fuel, position, hasLeaflet);
        }

        return instance;
    }

    public synchronized void setThingsInVision(List<Thing> list) {
        this.thingsInVision.clear();
        //hard copy
        for (Thing t : list) {
            thingsInVision.add(t);
        }
    }
    public synchronized List<Thing> getThingsInVision() {
        return this.thingsInVision;
    }

    public synchronized String getThingsNames() {
        String s = "";

        for (Thing t : thingsInVision) {
            s = s + " " + t.name;
        }
        return s;
    }

    
    public void setRobotID(String rid) {
        robotIndexID = rid;
    }


    public String getIndexID(){
        return this.robotIndexID;
    }
    public void update(String index, String colorName, double speed, double wheel, double pitch, double fuel, WorldPoint position, List<Leaflet> leafletList) { 
        this.color = translateIntoColor(colorName);
        if (speed == 0) {
            speed = Constants.DEFAULT_CREATURE_SPEED;//5 default (old: 1)
        }
        this.robotIndexID = index;
        this.a.speed = speed;
        this.a.wheel = wheel;
        this.s.pitch = pitch; //in RAD
        this.s.fuel = fuel;
        this.s.x = position.getX();
        this.s.y = position.getY();
        for(Leaflet l: leafletList){
            if (!ifHasLeaflet(l.getID())) {
                this.addLeaflet(l);
            }else{
                this.updateLeaflet(l.getID(), l.getItems());
            }
        }
    }

    public synchronized Creature updateState() {

        try {

            logger.info("----------- updateThingsStatus -----------");

            CommandUtility.updateStatus();


        } catch (CommandExecException ex) {
            java.util.logging.Logger.getLogger(WS3DProxy.class.getName()).log(Level.SEVERE, null, ex);
        }

        return this;
    }

    /**
     * Used in older versions.
     *
     * @param lo
     * @param angle
     * @param point
     */
    public void polarToXY(double lo, double angle, WorldPoint point) {
        double alpha;
        alpha = s.pitch - angle + Constants.M_PI_2;
        point.setX(s.x - lo * Math.cos(alpha));
        point.setY(s.y - lo * Math.sin(alpha));
    }

    /**
     * Return a bottom point of the frustrum far plane. Refer to "OpenGL frustru
     * culling" for more details.
     *
     * @param farDist
     * @param angle
     * @return
     */
    public WorldPoint getBottomPointFarPlane(double farDist, double angle) {
        double alpha, lo;
        //isosceles triangle:
        //side = 2*Height*sqrt(3)/3
        lo = (2 * farDist * Math.sqrt(3)) / 3;

        WorldPoint point = new WorldPoint();
        alpha = s.pitch - angle + Constants.M_PI_2;
        point.setX(s.x - lo * Math.cos(alpha));
        point.setY(s.y - lo * Math.sin(alpha));

        return point;
    }

    /**
     * Calculate the polar coordinates relative to the creature position
     *
     * @param x - the x cartesian coordinate
     * @param y - the y cartesian coordinate
     * @return double[] {distance, angle}
     */
    public double[] xyToPolar(double x, double y) {
        //cartesian coords:
        double dx = x - s.x;
        double dy = y - s.y;

        //kinematics model
        double lo = Math.sqrt(dx * dx + dy * dy); //ro
        //alpha - angle between the creature and the target
        double to = -s.pitch - Math.atan2(-dy, dx) + Constants.M_PI_2;

        //[-PI , PI]
        while (to > Constants.M_PI) {
            to -= 2 * Constants.M_PI;
        }

        while (to < -Constants.M_PI) {
            to += 2 * Constants.M_PI;
        }

        //in polar coords {distancia, angulo}
        return new double[]{lo, to};
    }

    public synchronized WorldMap getWorldMap() {
        return worldMap;
    }

    public synchronized void setWorldMap(WorldMap worldMap) {
        this.worldMap = worldMap;
    }

    public synchronized double getSpeed() {
        return this.a.speed;
    }

    public synchronized int getMotorSys() {
        return this.motorSys;
    }

    public synchronized double getWheel() {
        return this.a.wheel;
    }

    public synchronized double getPitch() {
        return this.s.pitch;
    }

    public synchronized double getFuel() {
        return this.s.fuel;
    }

    public synchronized WorldPoint getPosition() {
        WorldPoint p = new WorldPoint(s.x, s.y);
        logger.info(" Creature at: x= " + s.x + " and y= " + s.y);
        return p;
    }

    public Color getColor() {
        return this.color;
    }

    private Color translateIntoColor(String name) {
        return Constants.translateIntoColor(name);
    }

    public long getCreatureSize() {

        return this.mySize;
    }

    public boolean ifIsAt(WorldPoint point) {
        double pointX = point.getX();
        double pointY = point.getY();

        return (this.xyToPolar(pointX, pointY)[0] == 0);
    }

    public synchronized void addLeaflet(Leaflet l) {
        myLeaflets.put(l.getID(), l);
    }

    public synchronized Vector<Leaflet> getLeaflets() {
        Vector v = new Vector();
        for (Iterator iter = myLeaflets.values().iterator(); iter.hasNext();) {
            Leaflet l = (Leaflet) iter.next();
            v.add(l);
        }
        return v;

    }

    public synchronized boolean ifHasLeaflet(Long ID) {
        boolean ret = false;
        if (myLeaflets.containsKey(ID)) {
            ret = true;
        }
        return ret;
    }

    public synchronized void updateLeaflet(Long ID, HashMap items) {

        myLeaflets.get(ID).setItems(items);
    }

    public String getName() {
        return this.name;
    }

    /**
     * Refer to "OpenGL frustrum culling" for more details.
     *
     * @return
     */
    private double getHalfAngleFOVX() {
        World parameters = World.getInstance();

        //these settings must be according to those on the "server"
        double farDist = 1000;
        double ratio = parameters.getEnvironmentWidth() / parameters.getEnvironmentHeight();
        double fovY = 1.03; //~(pi/4 + err due to roundings)

        double Wfar = 2 * farDist * ratio * Math.tan(fovY / 2);
        double halfAlpha = Math.asin((Wfar / 2) / Math.sqrt((farDist * farDist) + (Wfar * Wfar) / 4));
        return halfAlpha;
    }

    /**
     * Generates a "triangule" that represents the field of view (horizontal
     * perspective) of the creature (what is seen on its camera).
     *
     * @param creature
     * @param g
     * @return
     */
    public Polygon getFOV() {
        double farDist = 1000; //settings from "server" (setFrustumPerspective)
        WorldPoint p = this.getPosition();
        double xCreaturePosition = p.getX();
        double yCreaturePosition = p.getY();
        Polygon pol = new Polygon();
        pol.addPoint((int) xCreaturePosition, (int) yCreaturePosition);

        double ang = getHalfAngleFOVX();
        p = this.getBottomPointFarPlane(farDist, ang);
        pol.addPoint((int) p.getX(), (int) p.getY());
        p = this.getBottomPointFarPlane(farDist, -ang);
        pol.addPoint((int) p.getX(), (int) p.getY());

        this.s.FOV = pol;
        return pol;
    }

    public synchronized int getIfHasLeaflet() {
        return this.hasLeaflet;
    }

    /**
     * Command to start the motor system of the creature.
     *
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void start() throws CommandExecException {
        CommandUtility.sendStartCreature(robotIndexID);
    }

    /**
     * Command to stop the motor system of the creature.
     *
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void stop() throws CommandExecException {
        CommandUtility.sendStopCreature(robotIndexID);
    }

    /**
     * Set the rotation around the creature's axis of symmetry. The creature's
     * movement is changed in order to make it faces towards a certain
     * direction. The creature speed (i.e. linear velocity) is the arithmetic
     * mean of Vr and Vl. The angle must be in rad.
     *
     * @param vr linear velocity of the right wheel
     * @param vl linear velocity of the left wheel
     * @param w the new pitch of the creature in rad
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void move(double vr, double vl, double w) throws CommandExecException {
        CommandUtility.sendSetAngle(robotIndexID, vr, vl, w);
    }

    /**
     * The creature grasp and put an object in its sack. The object may be: food
     * or crystal.
     *
     * @param thingName the unique name of the Thing. Example:
     * Jewel_1363904023537
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void putInSack(String thingName) throws CommandExecException {
        CommandUtility.sendPutInSack(robotIndexID, thingName);
    }

    /**
     * The creature eats a food.
     *
     * @param thingName the unique name of the Thing. Example:
     * Jewel_1363904023537
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void eatIt(String thingName) throws CommandExecException {
        CommandUtility.sendEatIt(robotIndexID, thingName);
    }

    /**
     * The creature grasp and hide an object under the ground.
     *
     * @param thingName the unique name of the Thing. Example:
     * Jewel_1363904023537
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void hideIt(String thingName) throws CommandExecException {
        CommandUtility.sendHideIt(robotIndexID, thingName);
    }

    /**
     * The creature digs an object up. The object must have been hided before.
     *
     * @see #sendHideIt
     *
     * @param thingName the unique name of the Thing. Example:
     * Jewel_1363904023537
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void unhideIt(String thingName) throws CommandExecException {
        CommandUtility.sendUnhideIt(robotIndexID, thingName);
    }

    /**
     * Starts the camera (visual sensor) of the creature.
     *
     * Note: Currently, there are only 2 cameras available on the server. The
     * right camera is attached to even robotIDs and the left camera, to the odd
     * robotIDs.
     *
     * @param robotID sequence number assigned to each object once it is
     * created.
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void startCamera(String robotID) throws CommandExecException {
        CommandUtility.sendStartCamera(robotID);
    }

    /**
     * Randomly generates a new leaflet. A leaflet is a list of crystals to be
     * collected and delivered at the DeliverySpot.
     *
     * The leaflets (3 in total). Format: "Leaflet1 Leaflet2 Leaflet3" Leaflet
     * format: "[Color TotalNumber NumberCollected] Payment Example: White 2 0
     * Yellow 1 0 6 White 1 0 Magenta 2 0 5 Blue 1 0 Red 1 0 Green 1 0 24
     * Leaflet1: 2 white jewels and 1 yellow jewel. None already collected. This
     * leaflet increases the score in 6 points Leaflet2: 1 white jewel and 2
     * magenta jewels. None already collected. This leaflet increases the score
     * in 5 points Leaflet3: 1 Blue, 1 Red and 1 green jewel. None already
     * collected. This leaflet increases the score in 24 points
     *
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void genLeaflet() throws CommandExecException {
        CommandUtility.sendGenLeaflet();
    }

    /**
     * The leaflet is delivered at the DeliverySpot. All the crystals specified
     * in the leaflet were collected and the creature is at the DeliverySpot.
     *
     * @throws CommandExecException An exception is thrown in case of missing or
     * invalid parameter
     */
    public synchronized void deliverLeaflet() throws CommandExecException {
        CommandUtility.sendDeliverLeaflet(robotIndexID);
    }

    /**
     * Check if a specific leaflet has not been assigned to the creature yet.
     * @param ID the leafletID
     * @return true: has not been assigned, so must do it; false otherwise
     */
    public synchronized boolean mustInitializeLeaflet(Long ID) {
        boolean ret = false;

        if (!ifHasLeaflet(ID)) {
            //the leaflet has not been created yet, create now.
            ret = true;
        }
        return ret;
    }
}
